/*
 * Exif ֘A
 */
#include "cpu.h"
#include "viewer.h"
#include "diskio.h"
#include "ff.h"
#include "monitor.h"
#include "exif.h"

/*
 * JPEG }[J`
 * {\tgł́ASOIȂтAPP1Ȃ
 * SOI̒APP1邱Ƃ҂邪AAPP0ꍇɂ
 * ǂݔ΂B
 */
#define	M_SOI	0xFFD8
#define	M_APP0	0xFFE0
#define	M_APP1	0xFFE1

typedef struct {
  int swap;
  UCHAR *tbuffer;
  UCHAR *readptr;
  char buffer[100];
} TFILE;

#define	XVERBOSE

long Process_IFD(TFILE *tfp, EXIF *exif);

short tiff_read_short(TFILE *tfp)
{
    short val;

    if (tfp->swap) {
       val = (tfp->readptr[1] << 8) | tfp->readptr[0];
    } else {
       val = (tfp->readptr[0] << 8) | tfp->readptr[1];
    }
    tfp->readptr += 2;
    return val;
}

long tiff_read_long(TFILE *tfp)
{
    long val;
    unsigned char *pbuff;

    pbuff = tfp->readptr;
    if (tfp->swap) {
       val = (pbuff[3] << 24) | (pbuff[2] << 16) | (pbuff[1] << 8) | pbuff[0];
    } else {
       val = (pbuff[0] << 24) | (pbuff[1] << 16) | (pbuff[2] << 8) | pbuff[3];
    }
    tfp->readptr += 4;
    return val;
}

#ifdef VERBOSE
char *tag_format[] = {
  "UNDEF", "unsigned byte", "ascii string", "unsigned short",
  "unsigned long", "unsigned rational", "signed byte",
  "undefined", "signed short", "signed long", "signed rational",
  "signed float", "double float"
};
#endif

void Decode_String(char *tagname, TFILE *tfp, int numdata, int data)
{
    UCHAR *old_ptr;

    if (numdata <= 4) {
	tfp->buffer[0] = (data >> 24) & 0xff;
	tfp->buffer[1] = (data >> 16) & 0xff;
	tfp->buffer[2] = (data >> 8) & 0xff;
	tfp->buffer[3] = data & 0xff;
	tfp->buffer[numdata] = 0;
#ifdef VERBOSE
	xprintf("\t%s: %s\n", tagname, tfp->buffer);
#endif
	return;
    }
    if (numdata < 100) {
	old_ptr = tfp->readptr;
 	tfp->readptr = tfp->tbuffer + data;
	memcpy(tfp->buffer, tfp->readptr, numdata);
	tfp->buffer[numdata] = 0;
#ifdef VERBOSE
	xprintf("\t%s: %s\n", tagname, tfp->buffer);
#endif
	tfp->readptr = old_ptr;
    }
}

void string_tag(char *pb, TFILE *tfp, int numdata, int data)
{
    UCHAR *old_ptr;

    if (numdata <= 4) {
	pb[0] = (data >> 24) & 0xff;
	pb[1] = (data >> 16) & 0xff;
	pb[2] = (data >> 8) & 0xff;
	pb[3] = data & 0xff;
	pb[numdata] = 0;
	return;
    }
    if (numdata > EXIF_STR_LEN)
        numdata = EXIF_STR_LEN;
    old_ptr = tfp->readptr;
    tfp->readptr = tfp->tbuffer + data;
    memcpy(pb, tfp->readptr, numdata);
    tfp->buffer[numdata] = 0;
    tfp->readptr = old_ptr;
}

void rational_tag(long *pval, TFILE *tfp, long offset)
{
    UCHAR *old_ptr;

    old_ptr = tfp->readptr;
    tfp->readptr = tfp->tbuffer + offset;
    pval[0] = tiff_read_long(tfp);
    pval[1] = tiff_read_long(tfp);
#ifdef DEBUG
    xprintf("%ld/%ld\n", pval[0], pval[1]);
#endif
    tfp->readptr = old_ptr;
}

void Decode_Rational(char *tagname, TFILE *tfp, long offset)
{
    UCHAR *old_ptr;
    long fact1, fact2;

    old_ptr = tfp->readptr;
    tfp->readptr = tfp->tbuffer + offset;
    fact1 = tiff_read_long(tfp);
    fact2 = tiff_read_long(tfp);
    xprintf("\t%s: %ld/%ld\n", tagname, fact1, fact2);
    tfp->readptr = old_ptr;
}

short short_tag(TFILE *tfp, long val)
{
    if (tfp->swap)
	return (short) val;
    return (short) (val >> 16);
}

void tag_decode(EXIF *exif, TFILE *tfp, unsigned short tag, short format, long numdata, long con_data)
{
    UCHAR *old_ptr;
    int val;

    switch (tag) {
    case 0x0103:
#ifdef VERBOSE
	xprintf("\tCompression: %x\n", short_tag(tfp, con_data));
#endif
        exif->compression = short_tag(tfp, con_data);
	break;
#ifdef VERBOSE
    case 0x010e:
	Decode_String("ImageDescription", tfp, numdata, con_data);
	break;
#endif
    case 0x010f:
	string_tag(exif->maker, tfp, numdata, con_data);
#ifdef VERBOSE
	xprintf("\tMaker %s\n", exif->maker);
#endif
	break;
    case 0x0110:
	string_tag(exif->model, tfp, numdata, con_data);
#ifdef VERBOSE
	xprintf("\tModel %s\n", exif->model);
#endif
	break;
    case 0x0112:
	exif->orientation = short_tag(tfp, con_data);
#ifdef VERBOSE
	xprintf("\tOrientation: %d\n", exif->orientation);
#endif
	break;
    case 0x11a:
	rational_tag(exif->resolution, tfp, con_data);
#ifdef VERBOSE
	xprintf("XResolution %d/%d", exif->resolution[0], exif->resolution[1]);
#endif
	break;
#ifdef VERBOSE
    case 0x11b:
	Decode_Rational("YResolution", tfp, con_data);
	break;
#endif
    case 0x128:
	exif->resolution_unit = short_tag(tfp, con_data);
#ifdef VERBOSE
	xprintf("\tResolutionUnit: %d\n", exif->resolution_unit);
#endif
	break;
#ifdef VERBOSE
    case 0x0132:
	Decode_String("DateTime", tfp, numdata, con_data);
	break;
#endif
    case 0x201:
#ifdef VERBOSE
	xprintf("\tJpegInterchangeFormat: %lx\n", con_data);
#endif
	exif->thumb_offset = con_data;
	break;
#ifdef VERBOSE
    case 0x202:
	xprintf("\tJpegInterchangeFormatLength: %ld\n", con_data);
	break;
    case 0x8298:
	Decode_String("Copyright", tfp, numdata, con_data);
	break;
#endif
    case 0x829a:
#ifdef VERBOSE
	Decode_Rational("ExposureTime", tfp, con_data);
#else
	rational_tag(exif->exp_time, tfp, con_data);
#endif
	break;
    case 0x829d:
#ifdef VERBOSE
	Decode_Rational("FNumber", tfp, con_data);
#else
	rational_tag(exif->f_number, tfp, con_data);
#endif
	break;
    case 0x8769:
	old_ptr = tfp->readptr;
	tfp->readptr = tfp->tbuffer + con_data;
#ifdef VERBOSE
	xprintf("\tExifIFDPointer: %lx\n", con_data);
	xprintf("Next IFD = %lx\n", Process_IFD(tfp, exif));
#else
	Process_IFD(tfp, exif);
#endif
	tfp->readptr = old_ptr;
	break;
    case 0x8822:
        exif->exp_program = short_tag(tfp, con_data);
#ifdef VERBOSE
	xprintf("\tExposureProgram: %d\n", exif->exp_program);
#endif
	break;
    case 0x8827:
        exif->iso_speed = short_tag(tfp, con_data);
#ifdef VERBOSE
	xprintf("\tISOSpeedRatings: %d\n", exif->iso_speed);
#endif
	break;
#ifdef VERBOSE
    case 0x9000:
	xprintf("\tExifVersion: %c%c%c%c\n",
		(char) (con_data >> 24) & 0xff, (char) (con_data >> 16) & 0xff,
		(char) (con_data >> 8) & 0xff, (char) con_data & 0xff);
	break;
#endif
    case 0x9003:
	Decode_String("DateTimeOriginal", tfp, numdata, con_data);
	memcpy(exif->date_time, tfp->buffer, 20);
	exif->date_time[20] = '\0';
	break;
#ifdef VERBOSE
    case 0x9004:
	Decode_String("DateTimeDigitized", tfp, numdata, con_data);
	break;
    case 0x9102:
	Decode_Rational("CompressedBietsPerPixel", tfp, con_data);
	break;
    case 0x9202:
	Decode_Rational("ApertureValue", tfp, con_data);
	break;
    case 0x9203:
	Decode_Rational("BrightnessValue", tfp, con_data);
	break;
#endif
    case 0x9207:
	exif->metering = short_tag(tfp, con_data);
#ifdef VERBOSE
	xprintf("\tMetringMode: %d\n", exif->metering);
#endif
	break;
    case 0x9208:
	exif->light_source = short_tag(tfp, con_data);
#ifdef VERBOSE
	xprintf("\tLightSource: %d\n", exif->light_source);
#endif
	break;
    case 0x9209:
        exif->flash = short_tag(tfp, con_data);
#ifdef VERBOSE
	xprintf("\tFlash: %d\n", exif->flash);
#endif
	break;
    case 0x920a:
#ifdef VERBOSE
	Decode_Rational("FocalLength", tfp, con_data);
#else
	rational_tag(exif->focal_length, tfp, con_data);
#endif
	break;
#ifdef VERBOSE
    case 0x927c:
	xprintf("\tMakerNote:\n");
	break;
    case 0xa001:
	xprintf("\tColorSpace: %d\n", short_tag(tfp, con_data));
	break;
#endif
    case 0xa002:
	val = con_data;
	if (format == 3) val = short_tag(tfp, con_data) & 0xffff;
#ifdef VERBOSE
	xprintf("\tExifImageWidth: %d\n", val);
#endif
	exif->image_width = val;
	break;
    case 0xa003:
	val = con_data;
	if (format == 3) val = short_tag(tfp, con_data) & 0xffff;
#ifdef VERBOSE
	xprintf("\tExifImageHeight: %d\n", val);
#endif
	exif->image_height = val;
	break;
    case 0xa402:	/* ExposureMode */
	exif->exp_mode = short_tag(tfp, con_data);
	break;
    case 0xa403:	/* WhiteBalance */
	exif->wb_mode = short_tag(tfp, con_data);
	break;
    }
}

long Process_IFD(TFILE *tfp, EXIF *exif)
{
    long link, num_data, con_data;
    unsigned short tag;
    short num_ent, format;

    num_ent = tiff_read_short(tfp);
#ifdef VERBOSE
    xprintf("IFD Number of entries: %d\n", num_ent);
#endif
    while (num_ent) {
	tag = (unsigned short) tiff_read_short(tfp);
	format = tiff_read_short(tfp);
	num_data = tiff_read_long(tfp);
	con_data = tiff_read_long(tfp);
#ifdef VERBOSE
        xprintf("  Tag: %04x, Num: %ld (%s)\n", tag,
		num_data, tag_format[format]);
#endif
	tag_decode(exif, tfp, tag, format, num_data, con_data);
	num_ent--;
    }
    link = tiff_read_long(tfp);
    return link;
}

/*
 * TIFF^ÕfR[h
 */
void process_tiff(UCHAR *tiff_top, EXIF *exif)
{
    int swap;
    short magic;
    long offset;
    TFILE tfinfo, *tfp;

    /* oCgI[_̊mF */
    if (tiff_top[0] == 'I' && tiff_top[1] == 'I') {
	/* Intel format. Need to swap */
        swap = 1;
    } else if (tiff_top[0] == 'M' && tiff_top[1] == 'M') {
	/* Motorola format. No need to swap */
	swap = 0;
    } else {
	xprintf("Bad Byte Align.\n");
	return;
    }
    tfp = &tfinfo;
    tfp->swap = swap;
    tfp->tbuffer = tiff_top;
    tfp->readptr = tiff_top + 2;
    magic = tiff_read_short(tfp);
    offset = tiff_read_long(tfp);
#ifdef DEBUG
    xprintf("swap = %d, magic = %x, offset = %lx\n", swap, magic, offset);
#endif
    while (offset) {
	tfp->readptr = tfp->tbuffer + offset;
	offset = Process_IFD(tfp, exif);
    }
}

/*
 * Exif ̓ǂݍ
 */
int CollectExif(FIL *fp, EXIF *exif)
{
    UCHAR buffer[12];
    USHORT app0_size, app1_size;
    UCHAR *tiff_top;
    FRESULT res;
    UINT cnt;

    /* 12oCg Exifwb_ǂݍ */
    res = f_read(fp, buffer, sizeof(buffer), &cnt);
    if (res != FR_OK || cnt != sizeof(buffer)) {
	xprintf("can't read Exif header.\n");
	return 0;
    }
    /* Exif  SOI}[JŎn܂Ă͂ */
    if (((buffer[0] << 8) | buffer[1]) != M_SOI) {
	xprintf("No SOI Marker.\n");
	return 0;
    }
    app0_size = 0;
    /* APP0 }[JA̕ǂݔ΂ */
    if (((buffer[2] << 8) | buffer[3]) == M_APP0) {
	app0_size = (buffer[4] << 8) | buffer[5];
	app0_size += 2;
	xprintf("APP0 found (size = %d)\n", app0_size);
	f_lseek(fp, app0_size + 2);
	f_read(fp, buffer + 2, sizeof(buffer)-2, &cnt);
    }
    /* APP1}[J邱ƂmF */
    if (((buffer[2] << 8) | buffer[3]) != M_APP1) {
	xprintf("No APP1 Marker.\n");
	exif->no_app1 = 1;
	return 0;
    }
    exif->no_app1 = 0;
    app1_size = (buffer[4] << 8) | buffer[5];

    /* Exif ̃}WbN邱ƂmF */
    if (memcmp(buffer + 6, "Exif\000\000", 6)) {
	xprintf("Bad exif header.\n");
	return 0;
    }
    pool_init();
    app1_size -= 8;
    /* APP1ǂݍ */
    tiff_top = (UCHAR *)malloc(app1_size);
    res = f_read(fp, tiff_top, app1_size, &cnt);
    if (res != FR_OK || cnt != app1_size) {
	xprintf("Can't read TIFF data part.\n");
	return 0;
    }
    /*
     * Tiff^O͂āǍʂi[
     */
    process_tiff(tiff_top, exif);
    exif->thumb_offset += sizeof(buffer) + app0_size;
#ifdef VERBOSE
    xprintf("thumb_offset = %d\n", exif->thumb_offset);
#endif
    return 1;
}
